/*
		MYASM v1.00							  (c) 1996 NikSoft International
		--------------------------------------------------------------------
		ASSEMBLE.H

		Header file for the module ASSEMBLE.C .

		The code generation modules are copied from NASM v0.90 the
		Netwide Assembler.

		10/11/96 18.50	Version v1.00
*/

#ifndef __ASSEMBLE_H
#define __ASSEMBLE_H

#include <stdio.h>

/* The data structure defining an output format driver, and the
   interfaces to the functions therein. */
typedef struct ofmt {
	/* This is a short (one-liner) description of the type of
	   output generated by the driver. */
	char *fullname;

	/* This is a single keyword used to select the driver. */
	char *shortname;

	/* This procedure is called at the start of an output session.
	   It tells the output format what file it will be writing to.
	   It also gives it a chance to do other initialisation. */
	int (*_init)(char *);

	/* This procedure is called by assemble() to write actual
	   generated code or data to the object file. Typically it
	   doesn't have to actually _write_ it, just store it for
	   later.

	   The `type' argument specifies the type of output data, and
	   usually the size as well: its contents are described below. */
	void (*_output)(long, long, void *, unsigned long, long, long);

	/* This procedure is called once for every symbol defined in
	   the module being assembled. It gives the name and value of
	   the symbol, in NASM's terms, and indicates whether it has
	   been declared to be global. Note that the parameter "name",
	   when passed, will point to a piece of static storage
	   allocated inside the label manager - it's safe to keep using
	   that pointer, because the label manager doesn't clean up
	   until after the output driver has.

	   Values of `is_global' are: 0 means the symbol is local; 1
	   means the symbol is global; 2 means the symbol is common (in
	   which case `offset' holds the _size_ of the variable).
	   Anything else is available for the output driver to use
	   internally. */
	void (*_symdef)(struct symbol *, int);

	/* This procedure is called when the source code requests a
	   segment change. It should return the corresponding segment
	   _number_ for the name, or NO_SEG if the name is not a valid
	   segment name.

	   It may also be called with NULL, in which case it is to
	   return the _default_ section number for starting assembly in.

	   It is allowed to modify the string it is given a pointer to. */
	long (*_section)(char *name);

	/* This procedure is called to modify the segment base values
	   returned from the SEG operator. It is given a segment base
	   value (i.e. a segment value with the low bit set), and is
	   required to produce in return a segment value which may be
	   different. It can map segment bases to absolute numbers by
	   means of returning SEG_ABS types. */
	long (*_segbase)(long segment);

	/* This procedure is called to allow the output driver to
	   process its own specific directives. When called, it has the
	   directive word in `directive' and the parameter string in
	   `value'. It is called in both assembly passes, and `pass'
	   will be either 1 or 2.

	   This procedure should return zero if it does not _recognise_
	   the directive, so that the main program can report an error.
	   If it recognises the directive but then has its own errors,
	   it should report them itself and then return non-zero. It
	   should also return non-zero if it correctly processes the
	   directive. */
	int (*_directive)(char *);

	/* This procedure is called after assembly finishes, to allow
	   the output driver to clean itself up and free its memory.
	   Typically, it will also be the point at which the object
	   file actually gets _written_.

	   One thing the cleanup routine should always do is to close
	   the output file pointer. */
	void (*_cleanup)();
} OFMT;

/* values for the `type' parameter to an output function. Each one
   must have the actual number of _bytes_ added to it.

   Exceptions are OUT_RELxADR, which denote an x-byte relocation
   which will be a relative jump. For this we need to know the
   distance in bytes from the start of the relocated record until
   the end of the containing instruction. _This_ is what is stored
   in the size part of the parameter, in this case.

   Also OUT_RESERVE denotes reservation of N bytes of BSS space,
   and the contents of the "data" parameter is irrelevant.

   The "data" parameter for the output function points to a "long",
   containing the address in question, unless the type is
   OUT_RAWDATA, in which case it points to an "unsigned char" array. */
#define OUT_RAWDATA					0x00000000UL
#define OUT_ADDRESS					0x10000000UL
#define OUT_REL2ADR					0x20000000UL
#define OUT_REL4ADR					0x30000000UL
#define OUT_RESERVE					0x40000000UL
#define OUT_TYPMASK					0xF0000000UL
#define OUT_SIZMASK					0x0FFFFFFFUL

/* Function for code generation. Through these pointers we can access the
   output driver functions */
extern int	(*DriverInit)(char *);
extern void (*Output)(long, long, void *, unsigned long, long, long);
extern int	(*Directive)(char *);
extern void (*Cleanup)();
extern void (*Symdef)(struct symbol *, int);
extern long (*Section)(char *);
extern long (*Segbase)(long);

extern long assemble(long, long, short, INSN *);
extern long instrsize(long, long, short, INSN *);
extern long pad(long, long, char, short);

/* Function for output format management */
extern int	SetFormat(char *name);
extern void FormatList();

#endif

/* --------- End of File --------------------------------------- */
