/*
	    JAS v1.00							  (c) 1996 NikSoft International
		--------------------------------------------------------------------
		BPATCH.C

		Backpatch handling.

		Performs the backpacth procedure for the symbol.

		Thursday, January 02, 1997	Version 1.00
		Saturday, May 10, 1997
			Added support for EXTOP backpatching.
*/

#include <stdlib.h>

#include "globals.h"
#include "ptypes.h"
#include "jaslib.h"
#include "symmngr.h"
#include "bpatch.h"
#include "assemble.h"

/* -------- Local Datas ----------------------------------------------- */

static short	UndefinedSymbols;
static SYMBOL	*UndefinedList;

/* -------- Local Functions' Prototypes ------------------------------- */

static void			DoBackPatch(PATCHREF *, long, long);

/* -------- Public Functions ------------------------------------------ */

/*
	Function......: BPatch
	Description...: Backpatches all the symbol forward references.
	Parameters....:
					Sym			pointer to the symbol to be patched.
*/
void BPatch(SYMBOL *Sym)
{
	register PATCHREF *Crsr = Sym->backpatchlist;
	register PATCHREF *Next;

	if ( !Crsr ) return;

	do {
		DoBackPatch(Crsr, Sym->offset, Sym->segment);

		/* Remove the patched reference */
		Next= Crsr->Next;
		LIB_free( Crsr );

		/* Move to the next reference */
	} while ( (Crsr= Next) );

	/* Decrement the counter for undefined symbols */
	UndefinedSymbols--;

	/* Remove the Undefined symbol from the 'unknown identifier' list */
	if ( Sym->UnknownPrev ) {
		Sym->UnknownPrev->UnknownNext= Sym->UnknownNext;
	} else {
		UndefinedList= Sym->UnknownNext;
	};

	if ( Sym->UnknownNext ) {
		Sym->UnknownNext->UnknownPrev= Sym->UnknownPrev;
	}
}


/*
	Function......: AddPatchRef
	Description...: Adds a backpatch reference to the backpatch list of
					the specified symbol.
	Parameters....:
					Sym				pointer to the symbol.
					PatchPtr		Pointer to the patch.
					Type			Patch type
					i1              Patch additional info.

*/
void AddPatchRef(SYMBOL *Sym, PATCH *PatchPtr, short Type, long i1)
{
	register PATCHREF *Ref = LIB_malloc(sizeof(PATCHREF));

	/* Add the reference to the head of the backpatch list of the symbol */
	if ( (Ref->Next = Sym->backpatchlist) ) {
		Ref->Next->Prev = Ref;
	} else {
		Sym->UnknownPrev= NULL;
		Sym->UnknownNext= UndefinedList;

		if ( UndefinedList != NULL ) {
			UndefinedList->UnknownPrev= Sym;
		};

		UndefinedList= Sym;
		UndefinedSymbols++;
	};

	Ref->Prev = NULL;
	Sym->backpatchlist= Ref;

	/* Another symbol referes to the PATCH */
	PatchPtr->RefCount++;

	/* Fill in some PATCHREF fields */
	Ref->ThePatch = PatchPtr;
	Ref->PatchType= Type;
	Ref->Occur	  = i1;
}

/*
	Function......: UndefSymNum
	Description...: Returns the number of symbols that are currently
					undefined, i.e. still to be backpatched.
	Parameters....: None.
*/
int UndefSymNum()
{
	SYMBOL *Crsr= UndefinedList;

	while ( Crsr ) {
		error(ERR_NONFATAL, UNDERR, Crsr->Name);
		Crsr= Crsr->UnknownNext;
	};

	return UndefinedSymbols;
}

/* -------- Local Functions ------------------------------------------- */

/*
	Function......: DoBackPatch
	Description...: Does the backpatch.
	Parameters....:
					Ref				pointer to the reference to patch.
					Offset			Offset of the patch.
					Segment			Segment of the patch.
*/
static void DoBackPatch(PATCHREF *Ref, long Offset, long Segment)
{
	register PATCH *aPatch= Ref->ThePatch;
	register INSN *PatchedInstr= &aPatch->patch_instr;
	unsigned long addr_size;

	aPatch->RefCount--;

	if ( Ref->PatchType < 0 || Ref->PatchType >= 3 ) {
		/* Bacpatch of a data definition */
		aPatch->patch_off += Offset * Ref->Occur;
		if ( aPatch->patch_seg == NO_SEG ) {
			aPatch->patch_seg= Segment;
		};

		/* This is the last reference ... */
		if ( aPatch->RefCount == 0 ) {
			/* ... so generate code for the patch ... */
			addr_size = OUT_ADDRESS + ( aPatch->Bits == 16 ? 2L : 4L );
			Output(aPatch->Seg, aPatch->Offset, &aPatch->patch_off,
				addr_size, aPatch->patch_seg, NO_SEG);

			/* ... and remove the patch. */
			LIB_free(aPatch);
		};
	} else {
		/* Backpatch of an operand */
		PatchedInstr->oprs[Ref->PatchType].offset += Offset * Ref->Occur;
		if ( PatchedInstr->oprs[Ref->PatchType].segment == NO_SEG ) {
			PatchedInstr->oprs[Ref->PatchType].segment = Segment;
		};

		/* If this is the last reference to be patched ... */
		if ( aPatch->RefCount == 0 ) {
			/* ... generate code for the instruction ...*/
			assemble(aPatch->Seg, aPatch->Offset, aPatch->Bits,
			    PatchedInstr);
			/* ... and remove the patch */
			LIB_free(aPatch);
		};
	};
}

/* -------- End Of File ----------------------------------------------- */
