/*
		JAS v1.00							  (c) 1997 NikSoft International
		--------------------------------------------------------------------
		EVAL.C

		Evaluates the expressions returned by the parser.

		Evaluator functions prototypes.
		Thursday, February 20, 1997 Version 1.00
			Nothing particular, simply the original static functions of
			the parser module are now here.
*/

#include <stdarg.h>

#include "globals.h"
#include "jaslib.h"
#include "eval.h"
#include "symmngr.h"
#include "bpatch.h"

#define EVALMEM_SIZE					1024

/* -------- Static Datas ---------------------------------------------- */

static EXPR_RESULT	EvaluatorMem[EVALMEM_SIZE];
static EXPR_RESULT	*CellPtr;
static short		flagCritical;

/* -------- Public Functions ------------------------------------------ */

/*
	Function......: SetCritical
	Description...: Set the state of the flagCritical.
	Parameters....: Obvious!
*/
void SetCritical(short flag)
{
	flagCritical= flag;
}


/*
	Function......: IsCritical
	Description...: Returns the state of the flagCritical.
	Parameters....: None.
*/
short IsCritical()
{
	return flagCritical;
}


/*
	Function......: EvaluatorReset
	Description...: Resets the evaluator.
					This function must be called before parsing an
					expression.
	Parameters....: None.
*/
void EvaluatorReset()
{
	CellPtr     = EvaluatorMem;
	flagCritical= 0;
}


/*
	Function......: IncCellPtr
	Description...: Increments the pointer to the evaluator memory to
					next cell. Returns the previous value of the pointer.
					If error are encountered (i.e. the evaluator is out of
					memory) it calls the global error reporting function.
	Parameters....: None.
*/
EXPR_RESULT *IncCellPtr()
{
	if ( CellPtr >= &EvaluatorMem[EVALMEM_SIZE] ) {
		error(ERR_FATAL, GENERR, "Formula too complex");
		return NULL;
	};

	return CellPtr++;
}


/*
	Function......: Vectorize
	Description...: Stores the specified data in a sequence of cells of
					the evaluator memory.
					Returns the address of the first cell of the sequence.
	Parameters....:
					Type			Type of the datas to be stored.
					Terminate		Flag to toggle expression termination.
					...			    The datas.
*/
EXPR_RESULT *Vectorize(int Type, int Terminate, ...)
{
	register EXPR_RESULT *RetVal= CellPtr;
	va_list arg_ptr;

	va_start(arg_ptr, Terminate);

	switch ( CellPtr->type = Type ) {
		case FLOAT:
			CellPtr->flt_result= va_arg(arg_ptr, float);
			break;

		case CHAR:
			CellPtr->chr_result= va_arg(arg_ptr, char);
			break;

		case STR:
			CellPtr->str_result= va_arg(arg_ptr, char *);
			CellPtr->str_leng  = va_arg(arg_ptr, int);
			break;

		case SYMREF:
			CellPtr->SymId	 = va_arg(arg_ptr, SYMBOL *);
			CellPtr->SymOccur= 1;
			break;

		case SEGMENT:
		case INTEGER:
			CellPtr->int_result= va_arg(arg_ptr, int);
			break;

		default:
			if ( Type <= 0 || Type >= MAXTYPE ) {
				error(ERR_FATAL, GENERR, "Evaluator error");
				va_end(arg_ptr);
				return NULL;
			};
			/* Vectorize a register */
			CellPtr->int_result= 1;
	};

	IncCellPtr();

	/* Terminate the expression if requested */
	if ( Terminate ) {
		CellPtr->type = EXPR_END;
	    IncCellPtr();
	};

	va_end(arg_ptr);
	return RetVal;
}


/*
	Function......: AddExprs
	Description...: Adds two vectors representing the expression.
	Parameters....:
					p, q		pointers to the vectors to be added.
*/
EXPR_RESULT *AddExprs(register EXPR_RESULT *p, register EXPR_RESULT *q)
{
	register EXPR_RESULT *r = CellPtr;

	while (p->type != EXPR_END && q->type != EXPR_END ) {
		if ( p->type > q->type ) {
	        *CellPtr = *q;
			q++;
		} else if (p->type < q->type) {
			*CellPtr = *p;
			p++;
		} else {		       /* *p and *q have same type */
	        switch ( CellPtr->type = p->type ) {
				case INTEGER:
					CellPtr->int_result= p->int_result + q->int_result;
					break;

				case SEGMENT:
					CellPtr->int_result= ( p->int_result <= q->int_result ?
						p->int_result : q->int_result );
					break;

				case FLOAT:
					CellPtr->flt_result= p->flt_result + q->flt_result;
					break;

				case CHAR:
					CellPtr->chr_result= p->chr_result + q->chr_result;
					break;

				case STR:
					CellPtr->str_result=
						LIB_strcat(p->str_result, q->str_result);
					CellPtr->str_leng  = p->str_leng + q->str_leng;
					LIB_free(p->str_result);
					LIB_free(q->str_result);
					break;

				case SYMREF:
					if ( p->SymId == q->SymId ) {
						CellPtr->SymId	 = (SYMBOL *)p->SymId;
						CellPtr->SymOccur= p->SymOccur + q->SymOccur;
					} else {
						*CellPtr = *p;
						IncCellPtr();
						*CellPtr = *q;
					};
					break;
			};

	        p++;
			q++;
		};

		IncCellPtr();
	}

	/* Copy the rest of the expression and terminate the result */
	if ( p->type == EXPR_END ) {
		do {
			*CellPtr = *q;
			IncCellPtr();
		} while ( q->type != EXPR_END && q++ );
	} else {
		do {
			*CellPtr = *p;
			IncCellPtr();
		} while ( p->type != EXPR_END && p++ );
	}

	return r;
}


/*
	Function......: ScalarMult
	Description...: Multiplies a vector expression by a scalar.
					Returns the pointer to the resulted expression or
					null on errors.
	Parameters....:
					scalar		scalar value to be multiplied.
					e			pointer to the expression.

*/
EXPR_RESULT *ScalarMult(long scalar, EXPR_RESULT *e)
{
	register EXPR_RESULT	*p= e;

	while ( p->type != EXPR_END ) {

		switch ( p->type ) {
			case FLOAT:
				p->flt_result *= (double)scalar;
				break;

			case SYMREF:
				p->SymOccur *= scalar;
				break;

			default:
			case INTEGER:
			case CHAR:
			/* Register integer or char */
				p->int_result *= scalar;

			case STR:
				break;
		}

		p++;
	};

	return e;
}


/*
	Function......: IsSimple
	Description...: Returns the type of the expression if it is
					simple, -1 otherwise.
	Parameters....:
					e			pointer to the expression to be
								analyzed. */
int IsSimple(register EXPR_RESULT *e)
{
	int RetVal = EXPR_END;

	for ( ;e->type != EXPR_END; e++ ) {
		/* Skip segment parts */
		if ( e->type == SEGMENT ) continue;

		if ( RetVal != e->type ) {
		/* Simple type cast mechanism */
			switch ( RetVal ) {
				case EXPR_END:
					RetVal= e->type;
					if ( RetVal == SYMREF ) RetVal= INTEGER;
					break;

				case INTEGER:
					switch ( e->type ) {
						case CHAR:
					    /* CHAR are upcasted to INTEGER */
						    e->type		  = INTEGER;
						    e->int_result = e->chr_result;

						case SYMREF:
							if ( flagCritical ) {
								error(ERR_NONFATAL, UNDERR, e->SymId->Name);
			            		FreeSymbol(e->SymId);
								return -1;
							};
							break;

						default:
						/* Incompatible types, cast is impossible */
						    return -1;
					};
					break;

				case FLOAT:
					if ( e->type == INTEGER ) {
					/* INTEGER are upcasted to FLOAT */
						RetVal		  = FLOAT;
						e->type		  = FLOAT;
						e->flt_result = (float)e->int_result;
					} else {
						return -1;
					};
					break;

				default:
					return -1;
			};
		};
	};

	return ( RetVal >= INTEGER ? RetVal : -1 );
}


/*
	Function......: FloatVal
	Description...: Returns the floating point part of an expression.
	Parameters....: See Above.
*/
float FloatVal(register EXPR_RESULT *e)
{
	float RetVal = 0.0;

	while ( e->type != EXPR_END ) {
		if ( e->type == FLOAT ) {
			RetVal += e->flt_result;
		};
		e++;
	};

	return RetVal;
}


/*
	Function......: SegmentVal
	Description...: Returns the segment part of an expression if any,
					NO_SEG otherwise.
	Parameters....: See Above. */
long SegmentVal(register EXPR_RESULT *e)
{
	while ( e->type != EXPR_END ) {
		if ( e->type == SEGMENT ) return e->int_result;
		e++;
	};

	return NO_SEG;
}


/*
	Function......: IntVal
	Description...: Same as the previous fuction but returns the integer
					part.
	Parameters....: See Above. */
int IntVal(register EXPR_RESULT *e)
{
	int RetVal = 0;

	while ( e->type != EXPR_END ) {
		if ( e->type == INTEGER ) {
			RetVal += e->int_result;
		};
		e++;
	};

	return RetVal;
}


/*
	Function......: ChrVal
	Description...: See above.
	Parameters....: See above. */
char ChrVal(register EXPR_RESULT *e)
{
	int RetVal = 0;

	while ( e->type != EXPR_END ) {
		if ( e->type == CHAR ) {
			RetVal += e->chr_result;
		};
		e++;
	};

	return (RetVal & 0x0ff);
}


/*
	Function......: StrVal
	Description...: See above.
	Parameters....: See above. */
char *StrVal(register EXPR_RESULT *e)
{
	char *RetVal = NULL;
	char *PrevVal = NULL;

	while ( e->type != EXPR_END ) {
		if ( e->type == STR ) {
			if ( !RetVal ) {
				RetVal = e->str_result;
			}
			else {
				PrevVal = RetVal;
				RetVal = LIB_strcat(RetVal, e->str_result);
				LIB_free(PrevVal);
			};
		};
		e++;
	};

	return RetVal;
}

/* -------- End Of File ----------------------------------------------- */
