/*
		JAS v1.00							  (c) 1996 NikSoft International
		--------------------------------------------------------------------
		JASMAIN.C

		Main module.
		Command line parsing and input file process.
		Global datas and functions.

		NOTE: Part of the JAS source code is copied form NASM v0.90, the
			  Netwide Assembler, in particular all the code generation
			  modules are taken from NASM.

		09/11/96 17.47	Version v1.00
		Thursday, February 20, 1997
			The original project is now called JAS The Jet ASsembler.
*/

#include "globals.h"
#include "jaslib.h"

extern int	Compile(char *, char *);
extern void FormatList();
extern int	SetDriver(char *);
extern int	yydebug;

/* --------- Global datas -------------------------------------- */

short	MaxErrors = -1;
short	IdCase	  = -1;
short	Errors	  = 0;

int	PoolPtr;
char ParamPool[2048];

/* --------- Static Datas -------------------------------------- */

static char InputFileName[MAX_FILENAME];
static char OutputFileName[MAX_FILENAME];
static char JASlogo[] =
	"JAS Jet ASsembler (c) 1997 NikSoft International.";

/* --------- Static functions ---------------------------------- */

/*
	Function....:	ParseCommandLine
	Description.:	Parse the command line.
					Return 0 if successfull, -1 and call the error
					reporting function if error are encountered.
					If the command line contains the options '-v' and
					'-l' the function returns -1 so the program terminates.
	Parameters..:
					argCount				Number of argument to parse.
					Args					Pointer to the array of args. */
static int ParseCommandLine(int argCount, char **Args)
{
	while ( argCount && strlen(*Args) == 2 && **Args == '-' ) {
		switch ( (*Args)[1] ) {
			case 'j':
				argCount--;
				Args++;
				if ( !*Args || sscanf(*Args, "%hu", &MaxErrors) != 1 ) {
					error(ERR_NONFATAL|ERR_USAGE|ERR_NOFILE,
						OPTPARAMERR, *Args);
					return -1;
				}
				argCount--;
				Args++;
				break;

			case 'v':
				puts(JASlogo);
				printf("Version v%s\n", JAS_VER);
				return -1;

			case 'l':
				puts("Output drivers list (* is the default).");
				FormatList();
				return -1;

			case 'o':
				argCount--;
				Args++;

				if ( !*Args ) {
					error(ERR_NONFATAL | ERR_USAGE | ERR_NOFILE,
						OPTPARAMERR, *Args);
					return -1;
				};

				strcpy(OutputFileName, *Args);
				argCount--;
				Args++;
				break;

			case 'f':
				argCount--;
				Args++;
				if ( SetDriver(*Args) ) {
					error(ERR_NONFATAL | ERR_NOFILE, FMTERR, *Args);
					return -1;
				};
				argCount--;
				Args++;
				break;

#ifndef __RELEASE
			case 'd':
				argCount--;
				Args++;
				yydebug = 1;
				break;
#endif

			case 'i':
				argCount--;
				Args++;
				IdCase= 0;
				break;

			default:
				error(ERR_NONFATAL|ERR_USAGE|ERR_NOFILE, OPTIONERR, *Args);
				return -1;
		}
	};

	/* The last argument must be the file name */
	if ( *Args ) {
		if ( **Args != '-' ) {
		    strcpy(InputFileName, *Args);
		    return 0;
		};
		error(ERR_NONFATAL|ERR_USAGE|ERR_NOFILE, OPTIONERR, *Args);
		return -1;
	};

	error(ERR_NONFATAL|ERR_NOFILE|ERR_USAGE, GENERR, "File name required");
	return -1;
}


/* --------- Function used by all modules ---------------------- */

/*
	Function......: getInputFileName
	Description...: Returns the pointer to the name of the input file.
	Parameters....: None. */
char *getInputFileName()
{
	return InputFileName;
}

/*
	Function......: getOuputFileName
	Description...: Returns the pointer to the name of the output file.
	Parameters....: None. */
char *getOutputFileName()
{
	return OutputFileName;
}

/* --------- Main ---------------------------------------------- */

void __crt0_load_environment_file(char *null)
{
}

char **__crt0_glob_function(char *null)
{
	return NULL;
}

int main(int argc, char *argv[])
{
	char *src;
	char *dst;

	/* Set the driver to the default */
	SetDriver( NULL );
	if ( ParseCommandLine(--argc, &argv[1]) ) return -1;

	/* Build up an output file name if none is specified. */
	if ( !*OutputFileName ) {
		src= InputFileName;
		dst= OutputFileName;

		while ( *src && *src != '.' ) *dst++ = *src++;
		*dst= '\0';
	}

	/* Parse the input file, assemble and generate output */
	if ( Compile(InputFileName, OutputFileName) ) return -1;

	return 0;
}

/* --------- End of File --------------------------------------- */
