/*
		JAS v1.00							  (c) 1996 NikSoft International
		--------------------------------------------------------------------
		OUTBIN.C

		Output routines for the Netwide Assembler to produce flat-form
		binary files.
		Code and general idea copied form NASM The NetWide Aseembler.

		The Netwide Assembler is copyright (C) 1996 Simon Tatham and
		Julian Hall. All rights reserved.
*/

#include "globals.h"
#include "jaslib.h"

#include "ptypes.h"
#include "parser.h"
#include "assemble.h"
#include "symmngr.h"

extern int yylex();		/* For format directive parsing */

#define TEXTSECT_DELTA							4096

/* -------------------------------------------------------------------- */

static FILE *fp;

static char *of_binerr[]= {
	"WRT not supported by binary output format",
	"Attempt to assemble code in [ABSOLUTE] space",
	"Attempt to initialise memory in the BSS section: ignored",
	"Binary output format does not support segment base references",
	"Binary output format does not support external references",
	"Uninitialised space not declared in BSS section: zeroing code or data",
	"binary output format does not support common variables"
};

static struct Section {
	unsigned char *contents;
	long length, size;
	long index;
} textsect, datasect;

static long bsslen, bssindex;

static struct Reloc {
	struct Reloc *next;
	long posn;
	long bytes;
	long secref;
	long secrel;
	struct Section *target;
} *relocs, **reloctail;


static int start_point;

static void make_room(struct Section *s, long bytes)
{
	if (s->length + bytes > s->size) {
		s->size     = s->length + bytes + TEXTSECT_DELTA;
		s->contents = LIB_realloc(s->contents, s->size);
	}
}

static void add_reloc(struct Section *s, long bytes, long secref, long offs,
	long secrel)
{
	struct Reloc *r;

	r = *reloctail = LIB_malloc(sizeof(struct Reloc));
	reloctail = &r->next;
	r->next = NULL;
	r->posn = offs;
	r->bytes = bytes;
	r->secref = secref;
	r->secrel = secrel;
	r->target = s;
}

static int bin_init(char *filename)
{
	if ( *filename ) {
		LIB_addExtension(filename, filename, ".bin");
	} else {
		strcpy(filename, "out.bin");
	};

	if ( !(fp= LIB_fopen(filename, M_RWBIN)) ) return -1;

	start_point = 0;		       /* default */
	textsect.contents = datasect.contents = NULL;
	textsect.length = textsect.size = datasect.length = datasect.size = 0;
	textsect.index = seg_alloc();
	datasect.index = seg_alloc();
	bsslen = 0;
	bssindex = seg_alloc();
	relocs = NULL;
	reloctail = &relocs;

	return 0;
}

static void bin_cleanup()
{
	struct Reloc *r;
	long datapos, dataalign, bsspos;

	datapos = (start_point + textsect.length + 3) & ~3;/* align on 4 bytes */
	dataalign = datapos - (start_point + textsect.length);

	bsspos = (datapos + datasect.length + 3) & ~3;

	for (r = relocs; r; r = r->next) {
	unsigned char *p, *q;
	long l;

	p = q = r->target->contents + r->posn;
	l = *p++;
	l += ((long)*p++) << 8;
	if (r->bytes == 4) {
	    l += ((long)*p++) << 16;
	    l += ((long)*p++) << 24;
	}

	if (r->secref == textsect.index) l += start_point;
	else if (r->secref == datasect.index) l += datapos;
	else if (r->secref == bssindex) l += bsspos;

	if (r->secrel == textsect.index) l -= start_point;
	else if (r->secrel == datasect.index) l -= datapos;
	else if (r->secrel == bssindex) l -= bsspos;

	if (r->bytes == 4)
	    WRITELONG(q, l);
	else
	    WRITESHORT(q, l);
	}

	fwrite(textsect.contents, textsect.length, 1, fp);
	if (datasect.length > 0) {
		fwrite("\0\0\0\0", dataalign, 1, fp);
		fwrite(datasect.contents, datasect.length, 1, fp);
	}
	fclose(fp);
	LIB_free(textsect.contents);
	LIB_free(datasect.contents);

	while (relocs) {
		r = relocs->next;
		LIB_free(relocs);
		relocs = r;
	}
}

static void bin_out(long segto, long offto, void *data, unsigned long type,
		     long segment, long wrt) {
	unsigned char *p;
	struct Section *s;
	long realbytes;

	if (wrt != NO_SEG) {
	    wrt = NO_SEG;		       /* continue to do _something_ */
	    error(ERR_NONFATAL, GENERR, of_binerr[0]);
	}

	/* handle absolute-assembly (structure definitions) */
	if (segto == NO_SEG) {
	    if ( (type & OUT_TYPMASK) != OUT_RESERVE )
	        error(ERR_NONFATAL, GENERR, of_binerr[1]);
	    return;
	}

	if (segto == bssindex) {	       /* BSS */
		if ( (type & OUT_TYPMASK) != OUT_RESERVE )
		    error(ERR_WARNING, GENERR, of_binerr[2]);
		s = NULL;
	} else if (segto == textsect.index) {
		s = &textsect;
	} else if (segto == datasect.index) {
		s = &datasect;
	}
	else {
		s = &textsect;
	};

	if ( (type & OUT_TYPMASK) == OUT_ADDRESS ) {
		if (segment != NO_SEG && segment != textsect.index &&
		    segment != datasect.index && segment != bssindex) {

		    if ( segment % 2 )
				error(ERR_NONFATAL, GENERR, of_binerr[3]);
	        else
				error(ERR_NONFATAL, GENERR, of_binerr[4]);

	        segment = NO_SEG;
	    }
		if (s) {
			if ( s->length >= offto ) {
		        make_room(s, type & OUT_SIZMASK);
		        if ( s->length == offto ) s->length += type & OUT_SIZMASK;
			} else {
				make_room(s, (type & OUT_SIZMASK) + (offto - s->length));
				s->length = offto + (type & OUT_SIZMASK);
			}

	        p = s->contents + offto;

	        if ( segment != NO_SEG )
		        add_reloc(s, type & OUT_SIZMASK, segment, offto, -1L);

		    if ((type & OUT_SIZMASK) == 4)
			    WRITELONG(p, *(long *)data);
		    else
			    WRITESHORT(p, *(long *)data);
		} else {
		    bsslen += type & OUT_SIZMASK;
		};
	} else if ( (type & OUT_TYPMASK) == OUT_RAWDATA ) {
		type &= OUT_SIZMASK;
		p = data;
		if (s) {
			if ( s->length >= offto ) {
		        make_room(s, type);
		        if ( s->length == offto ) s->length += type;
			}
			else {
				make_room(s, type + (offto - s->length));
				s->length = offto + type;
			};
			memcpy(s->contents + offto, data, type);
		}
		else {
		    bsslen += type;
		};
	} else if ( (type & OUT_TYPMASK) == OUT_RESERVE ) {
		type &= OUT_SIZMASK;
		if (s) {
		    error(ERR_WARNING, GENERR, of_binerr[5]);
		    make_room(s, type);
		    memset(s->contents + s->length, 0, type);
		    s->length += type;
		}
		else {
		    bsslen += type;
		};
	} else if ((type & OUT_TYPMASK) == OUT_REL2ADR ||
	       (type & OUT_TYPMASK) == OUT_REL4ADR) {
		realbytes = ((type & OUT_TYPMASK) == OUT_REL4ADR ? 4 : 2);
		if ( segment != NO_SEG && segment != textsect.index &&
		     segment != datasect.index && segment != bssindex) {

		    if ( segment % 2 )
				error(ERR_NONFATAL, GENERR, of_binerr[3]);
	        else
				error(ERR_NONFATAL, GENERR, of_binerr[4]);
		    segment = NO_SEG;
		}
		if (s) {
			if ( s->length >= offto ) {
		        make_room(s, realbytes);
		        if ( s->length == offto ) s->length += realbytes;
			}
			else {
				make_room(s, realbytes + (offto - s->length));
				s->length = offto + realbytes;
			};
		    p = s->contents + offto;
			add_reloc(s, realbytes, segment, offto, segto);

		    if ( realbytes == 4 )
			    WRITELONG(p, *(long *)data - realbytes - offto);
		    else
			    WRITESHORT(p, *(long *)data - realbytes - offto);
		}
		else {
		    bsslen += realbytes;
		};
	}
}

static void bin_deflabel(struct symbol *Sym, int is_global)
{
	if ( is_global == 2 ) {
	    error (ERR_NONFATAL, GENERR, of_binerr[6]);
	}
}

static long bin_secname(char *name)
{
	if ( !name ) return textsect.index;

	if ( !LIB_stricmp(name, "TEXT") ) return textsect.index;
	else if ( !LIB_stricmp(name, "DATA") ) return datasect.index;
	else if ( !LIB_stricmp(name, "BSS") ) return bssindex;
	else return NO_SEG;
}

static long bin_segbase (long segment) {
	return segment;
}

static int bin_directive(char *directive)
{
	if ( !strcmp(directive, "ORG") ) {

		if ( yylex() == TOKEN_NUM ) {
	        start_point = yylval.intval;
			return 0;
		};

		error(ERR_NONFATAL, INTERR);
	};

	error(ERR_NONFATAL, INVDIREC);
	return -1;
}

struct ofmt of_bin = {
	"flat-form binary files (e.g. DOS .COM, .SYS)",
	"bin",
	bin_init,
	bin_out,
	bin_deflabel,
	bin_secname,
	bin_segbase,
	bin_directive,
	bin_cleanup
};

/* -------- End Of File ----------------------------------------------- */
