; --------------------------------------------------------------------
; n6502 cpu emulator
; copyright 1997 Bloodlust Software
; free for public use
;
; --------------------------------------------------------------------
; Converted to JAS assembly language by Nicola Gaggi
; email:	gaggi@cs.unibo.it
; --------------------------------------------------------------------

.GLOBAL _ncpu
.GLOBAL _n6502_execute
.GLOBAL _n6502_int
.GLOBAL _n6502_nmi

.define F_CARRY			01h
.define F_ZERO			02h
.define F_INT			04h
.define F_DEC			08h
.define F_BREAK			10h
.define F_RSRVD			20h
.define F_OVER			40h
.define F_SIGN			80h

; .define MSVC


	  .SECTION DATA

;------------------------------------
;degenerate flagtable

flagtable:
.repeat 100h
  db (&count AND F_ZERO) XOR F_ZERO
  db ((&count AND F_OVER) SHR 6)
  db (&count AND F_CARRY)
  db (&count AND F_SIGN)
.end

	.SECTION BSS
;------------------------------------
;cpu struc
; JAS doesn't support structs yet!
n6502cpu:
n6502cpu_rA:	resb 1
n6502cpu_rS:	resb 1
n6502cpu_rX:	resb 1
n6502cpu_rY:	resb 1

n6502cpu_rF:	resb 1		; flags
n6502cpu_rPC:	resw 1		; PC
n6502cpu_trapbadops:	resb 3

n6502cpu_degenflags:
n6502cpu_zero:	 resb 1		; degenerate flags
n6502cpu_over:	 resb 1

n6502cpu_carry:	 resb 1		; (keep zero & sign in diff dwords)
n6502cpu_sign:	 resb 3

n6502cpu_cycles:		resd 1
n6502cpu_totalcycles:	resd 1

n6502cpu_PCBASE:		resd 1
n6502cpu_ROM:			resd 1	; pointer to ROM at 8000-FFFF
n6502cpu_RAM:			resd 1	; pointer to RAM at 0-7FF

n6502cpu_readtrap:		resd 1
n6502cpu_writetrap:		resd 1

n6502cpu_breakpoint:	resd 1


	  .SECTION TEXT


;------------------------------------
;fetch macros

.MACRO fetchbyte reg
	 mov	 &reg, byte [esi]
	 inc  esi
.END

.MACRO fetchbytesx reg
	 movsx	 &reg, byte [esi]
	 inc  esi
.END

.MACRO fetchbytezx reg
	 movzx	 &reg, byte [esi]
	 inc  esi
.END

.MACRO fetchword reg
	 mov	 &reg, word [esi]
	 add  esi,2
.END

.MACRO npushb reg
   mov [ebp + edx + 100h], &reg
   dec dl
.END

.MACRO npopb reg
   inc dl
   mov &reg, [ebp + edx + 100h]
.END


;-----------------------------
; PC nat/denat
; these must be customized

;converts 16-bit PC to a 32-bit base+PC in ESI
naturalizeESI PROC
   cmp esi, 8000h ;upper rom?
   jb  @@NOTROM
   mov [n6502cpu_PCBASE], edi
   add esi,edi
   ret

@@NOTROM:
   cmp esi,2000h ;mirrored ram?
   jae @@NOTRAM
   and esi,7FFh	 ;mirror lower 2k
@@NOTRAM:
   mov [n6502cpu_PCBASE],ebp
   add esi,ebp
   ret
naturalizeESI ENDP

;converts 32-bit base+PC to 16-bit PC
.MACRO denaturalizeESI
   sub esi, [n6502cpu_PCBASE]
.END


;--------------------------------
; read/write memory functions
; these must be customized

readbyte PROC
   cmp edx,8000h
   jb  @@NOTROM
   mov al,[edi+edx]
   ret
@@NOTROM:
   cmp edx,2000h ;mirrored ram?
   jae @@NOTRAM
   and edx,7FFh	 ;mirror lower 2k
   mov al, [ebp+edx]
   ret

@@NOTRAM:
   push ebx
   push ecx
   push edx
   push esi
   push edi

.if defined(MSVC)
	mov ecx, edx
.else
	mov eax, edx
.end

   mov	eax, edx
   call [n6502cpu_readtrap]
   pop	edi
   pop	esi
   pop	edx
   pop	ecx
   pop	ebx
   and	eax,0FFh
   ret
readbyte ENDP

writebyte PROC
   cmp edx,2000h
   jae @@NOTRAM
   and edx,7FFh	 ;mirror lower 2k
   mov [ebp+edx],al
   ret

@@NOTRAM:
   push ebx
   push ecx
   push edx
   push esi
   push edi

   call [n6502cpu_writetrap]
   pop	edi
   pop	esi

   test eax,eax
   jz	@@NOBANK
   ; reset ESI
   denaturalizeESI
   mov	edi,[n6502cpu_ROM]
   call naturalizeESI
@@NOBANK:

   pop	edx
   pop	ecx
   pop	ebx
   xor	eax,eax
   ret
writebyte ENDP




;----------------------------
; flag store macros

;saves zero,sign flags
.MACRO savef reg
   mov [n6502cpu_zero], &reg
   mov [n6502cpu_sign], &reg
.END

;saves zero,sign,carry flags
.MACRO savefc reg
   setc [n6502cpu_carry]
   mov [n6502cpu_zero], &reg
   mov [n6502cpu_sign], &reg
.END

;saves zero,sign,overflow flags
.MACRO savefo reg
   seto [n6502cpu_over]
   mov [n6502cpu_zero], &reg
   mov [n6502cpu_sign], &reg
.END

;saves zero,sign,carry,overflow flags
.MACRO savefco reg
   setc [n6502cpu_carry]
   seto [n6502cpu_over]
   mov [n6502cpu_zero], &reg
   mov [n6502cpu_sign], &reg
.END

;gets carry flag (Assumes ah=0)
.MACRO getc
	 cmp ah, [n6502cpu_carry]
.END


;-------------------------------------
; flag conversion macros

;al -> lastval,carry,over
.MACRO degenerateflags
   mov [n6502cpu_rF],al ;store full flags
   mov eax, [flagtable+eax*4]
   mov dword [n6502cpu_degenflags], eax
   xor eax,eax
.END

;lastval,carry,over -> al
.MACRO generateflags
   mov al, [n6502cpu_rF] ; al=RSRVD|DEC|INT|BREAK
   push edx

   and al, 3Ch
   mov dl, [n6502cpu_over]

   cmp	 ah, [n6502cpu_zero]
   jne	%%NOTZERO
   or	al,F_ZERO  ;set zero flag
%%NOTZERO:

   mov	ah, [n6502cpu_sign]
   and	dl,1

   shl	dl, 6
   and	ah, F_SIGN

   or	al, BYTE [n6502cpu_carry]  ;carry=bit 0
   or	dl, ah

   or	al, dl			  ;over=bit 7
   pop	edx

   xor ah,ah
.END


;-------------------
; execute function

_n6502_execute PROC USES esi,edi,ebx

	  xor	eax,eax
	  xor	esi,esi
	  mov	edi, [n6502cpu_ROM]
	  mov	ebp, [n6502cpu_RAM]
	  mov	bx, word [n6502cpu_rA]
	  mov	cx, word [n6502cpu_rX]
	  mov	si, [n6502cpu_rPC]

	  mov	al, [n6502cpu_rF] ;get flags
	  degenerateflags

	  call	naturalizeESI

	  xor	edx,edx
	  xor	eax,eax

executeloop:
	  sub	[n6502cpu_cycles], eax ;3
	  jle	done

	  mov	al, [esi]		   ;1
	  inc	esi

	  xor	edx, edx  ;agi

	  jmp	[n6502jmptable+eax*4] ;5


;	   mov	 edx,[n6502cpu_cycles] ;1
;	   sub	 edx,eax		  ;1
;	   mov	 al,[esi]
;	   mov	 [n6502cpu_cycles],edx ;1
;	   jle	 done
;	   xor	 edx,edx		   ;1
;	   inc	 esi
;	   jmp	[n6502jmptable+eax*4] ;5


badopcode:
	  mov	eax,4				            ; nop cycles
	  cmp	WORD [n6502cpu_trapbadops], 0   ; continue anyway?
	  jne	executeloop

	  dec	esi ;decrease PC ;abort
	  mov	eax, -1
	  jmp	short exit

done:
	  xor	eax,eax

exit:
	  sub	esi,[n6502cpu_PCBASE]
	  mov	word [n6502cpu_rA],bx
	  mov	word [n6502cpu_rX],cx
	  mov	[n6502cpu_rPC],si ;store PC

	  push eax
	  xor  eax,eax
	  ; set flags
	  generateflags
	  mov  [n6502cpu_rF],al
	  pop  eax

	  ret
_n6502_execute ENDP

;-----------------------------------
; interrupts


_n6502_nmi PROC USES edx, edi

   xor edx,edx
   xor eax,eax

   mov ebp, [n6502cpu_RAM]
   mov edi, [n6502cpu_ROM]

   mov ax, [n6502cpu_rPC] ;get pc
   mov dl, [n6502cpu_rS] ;get stack

   ;push pc
   npushb ah
   npushb al

   ;push flags
   mov al,[n6502cpu_rF]
   and al,NOT F_BREAK ;remove break flag ?
   npushb al

   ;store S
   mov [n6502cpu_rS],dl

   ;turn on interrupt flag
   or  al,F_INT
   mov [n6502cpu_rF],al

   ;PC=NMI vector
   mov ax,[edi+0FFFAh]
   mov WORD [n6502cpu_rPC], ax

   ret
_n6502_nmi ENDP



_n6502_int PROC USES edx, edi
   test BYTE [n6502cpu_rF], F_INT
   jnz	@@DONE

   xor edx,edx
   xor eax,eax

   mov ebp,[n6502cpu_RAM]
   mov edi,[n6502cpu_RAM]
   mov dl,[n6502cpu_rS] ;get stack
   mov ax,[n6502cpu_rPC] ;get pc

   ;push pc
   npushb ah
   npushb al

   ;push flags
   mov al,[n6502cpu_rF]
   and al,NOT F_BREAK ;remove break flag ?
   npushb al

   ;store S
   mov [n6502cpu_rS],dl

   or  al,F_INT ;turn on interrupt flag
   mov [n6502cpu_rF],al

   ;PC=INT vector
   mov edi,[n6502cpu_ROM]
   mov ax,[edi+0FFFEh]
   mov [n6502cpu_rPC],ax

@@DONE:
   ret
_n6502_int ENDP

;--------------------------------------
; opcode header/footer macros

.MACRO opbegin O
  .align	4
  op&O PROC
.END


.MACRO opend C
   mov al, &C ; store cycles
   jmp executeloop
   ENDP
.END

.MACRO opendzx C
   mov eax, &C ; store cycles
   jmp executeloop
   ENDP
.END

;---------------------------

.define nX		cl
.define nY		ch
.define nA		bl
.define nS		bh


;---------------------------------
; lea macros

;$XX
.MACRO lea_zero
	fetchbyte dl
.END

;$XX,x
.MACRO lea_zeroX
	fetchbyte dl
	add dl, nX
.END

;$XX,y
.MACRO lea_zeroY
	fetchbyte dl
	add dl,nY
.END

;$XXXX
.MACRO lea_abs
	fetchword dx
.END

;$XXXX,x
.MACRO lea_absX
	fetchword dx
	add	 dl, nX
	adc	 dh, ah ;ah=0
.END

;$XXXX,y
.MACRO lea_absY
	fetchword dx
	add	 dl,nY
	adc	 dh,ah ;ah=0
.END


;($XX,x)
.MACRO lea_indX
	fetchbyte dl
	add dl, nX
	mov dx, WORD [ebp+edx]
.END

;($XX),y
.MACRO lea_indY
	fetchbyte dl
	mov	 dx, WORD [ebp+edx]
	add	 dl,nY
	adc	 dh,ah	;ah=0
.END


;--------------------------------------
; read macros


;$XX
.MACRO Rzero
	lea_zero
	mov al, BYTE [ebp+edx]
.END

;$XX,x
.MACRO RzeroX
	lea_zeroX
	mov al, BYTE [ebp+edx]
.END

;$XX,y
.MACRO RzeroY
	lea_zeroY
	mov al, BYTE [ebp+edx]
.END

;$XXXX
.MACRO Rabs
	lea_abs
	call readbyte
.END

;$XXXX,x
.MACRO RabsX
	lea_absX
	call readbyte
.END

;$XXXX,y
.MACRO RabsY
	lea_absY
	call readbyte
.END


;($XX,x)
.MACRO RindX
	lea_indX
	call readbyte
.END

;($XX),y
.MACRO RindY
	lea_indY
	call readbyte
.END


;#$XX
.MACRO Rimmb
	fetchbyte al
.END



;-------------------------------------------
; write macros

;$XX
.MACRO Wzero
	mov BYTE [ebp+edx], al
.END

;$XX,x
.MACRO WzeroX
	mov BYTE [ebp+edx], al
.END

;$XX,y
.MACRO WzeroY
	mov BYTE [ebp+edx], al
.END

;$XXXX
.MACRO Wabs
	call writebyte
.END

;$XXXX,x
.MACRO WabsX
	call writebyte
.END

;$XXXX,y
.MACRO WabsY
	call writebyte
.END

;($XX,x)
.MACRO WindX
	call writebyte
.END

;($XX),y
.MACRO WindY
	call writebyte
.END


;------------------------------------
;opcodes

;NOP
opbegin _234
opend 2


;----------------
; bit/logic/arithmetic ops

;nA=nA+al+carry (decimal)
DECADC proc
   getc
   adc al,nA
   daa		   ;cheap, probably incorrect
   savefco al
   mov nA,al
   opend 4

;nA=nA-al+carry (decimal)
DECSBC proc
   getc
   cmc
   sbb nA,al
   mov al,nA
   das		   ;cheap, probably incorrect
   cmc
   savefco al
   mov nA,al
  opend 4

.MACRO ADC_op amode,op,cy
  opbegin &op
   R&amode
   test BYTE [n6502cpu_rF], F_DEC
   jnz	DECADC
   getc
   adc nA, al
   savefco nA
  opend &cy
.END

.MACRO SBC_op amode,op,cy
  opbegin &op
   R&amode
   test BYTE [n6502cpu_rF], F_DEC
   jnz	DECSBC
   xor	al,0FFh
   getc
   adc nA,al
   savefco nA
  opend &cy
.END

ADC_op	immb,_105,2
ADC_op	zero,_101,3
ADC_op zeroX,_117,4
ADC_op	 abs,_109,4
ADC_op	absX,_125,4
ADC_op	absY,_121,4
ADC_op	indX,_97,6
ADC_op	indY,_113,5

SBC_op	immb,_233,2
SBC_op	zero,_229,2
SBC_op zeroX,_245,4
SBC_op	 abs,_237,4
SBC_op	absX,_253,4
SBC_op	absY,_249,4
SBC_op	indX,_225,6
SBC_op	indY,_241,5


.MACRO LOG_op operator, amode, op, cy
  opbegin &op
	R&amode
	&operator nA, al
	savef nA
  opend &cy
.END

; AND
LOG_op and, immb,_41,2
LOG_op and, zero,_37,3
LOG_op and,zeroX,_53,4
LOG_op and,	 abs,_45,4
LOG_op and, absX,_61,4
LOG_op and, absY,_57,4
LOG_op and, indX,_33,6
LOG_op and, indY,_49,5

; OR
LOG_op	or, immb,_9,2
LOG_op	or, zero,_5,3
LOG_op	or,zeroX,_21,4
LOG_op	or,	 abs,_13,4
LOG_op	or, absX,_29,4
LOG_op	or, absY,_25,4
LOG_op	or, indX,_1,6
LOG_op	or, indY,_17,5

;EOR
LOG_op xor, immb,_73,2
LOG_op xor, zero,_69,3
LOG_op xor,zeroX,_85,4
LOG_op xor,	 abs,_77,4
LOG_op xor, absX,_93,4
LOG_op xor, absY,_89,4
LOG_op xor, indX,_65,6
LOG_op xor, indY,_81,5


.MACRO BIT_op amode,op,cy
  opbegin &op
	R&amode
	mov	  dl,al

	shr	  al,6
	mov	  [n6502cpu_sign],dl ;sign=mem

	and	  dl,nA		;acc & mem
	and	  al,1		;isolate bit 40h

	mov	  [n6502cpu_zero],dl ;set zero
	mov	  [n6502cpu_over],al
  opend &cy
.END

;BIT
BIT_op	 zero,_36,3
BIT_op	  abs,_44,4


;------------
;load/store


.MACRO LDx_OP reg,amode,op,cy
  opbegin &op
	R&amode
	mov &reg, al
	savef al
  opend &cy
.END

;LDA a
LDx_OP nA, immb,_169,2
LDx_OP nA, zero,_165,3
LDx_OP nA,zeroX,_181,4
LDx_OP nA,	abs,_173,4
LDx_OP nA, absX,_189,4
LDx_OP nA, absY,_185,4
LDx_OP nA, indX,_161,6
LDx_OP nA, indY,_177,5

;LDX a
LDx_OP nX, immb,_162,2
LDx_OP nX, zero,_166,3
LDx_OP nX,zeroY,_182,4
LDx_OP nX,	abs,_174,4
LDx_OP nX, absY,_190,4

;LDY a
LDx_OP nY, immb,_160,2
LDx_OP nY, zero,_164,3
LDx_OP nY,zeroX,_180,4
LDx_OP nY,	abs,_172,4
LDx_OP nY, absX,_188,4


.MACRO STx_OP reg,amode,op,cy
  opbegin &op
	lea_&amode
	mov	 al, &reg
	W&amode
  opend &cy
.END


;STA a
STx_OP nA, zero,_133,3
STx_OP nA,zeroX,_149,4
STx_OP nA,	abs,_141,4
STx_OP nA, absX,_157,5
STx_OP nA, absY,_153,5
STx_OP nA, indX,_129,6
STx_OP nA, indY,_145,6

;STX a
STx_OP nX, zero,_134,3
STx_OP nX,zeroY,_150,4
STx_OP nX, abs ,_142,4

;STY a
STx_OP nY, zero,_132,3
STx_OP nY,zeroX,_148,4
STx_OP nY, abs ,_140,4



;--------------------------------
; compare


.MACRO CMP_op reg,amode,op,cy
  opbegin &op
	R&amode
	mov ah, &reg
	sub ah, al
	setnc [n6502cpu_carry]
	mov	  [n6502cpu_sign],ah
	mov	  [n6502cpu_zero],ah
  opendzx &cy
.END


;CMP a
CMP_op nA, immb,_201,2
CMP_op nA, zero,_197,3
CMP_op nA,zeroX,_213,4
CMP_op nA,	abs,_205,4
CMP_op nA, absX,_221,4
CMP_op nA, absY,_217,4
CMP_op nA, indX,_193,6
CMP_op nA, indY,_209,5


;CPX a
CMP_op nX, immb,_224,2
CMP_op nX, zero,_228,3
CMP_op nX,	abs,_236,4

;CPY a
CMP_op nY, immb,_192,2
CMP_op nY, zero,_196,3
CMP_op nY,	abs,_204,4



;-------------------------------
;INC/DEC

;INX
opbegin _232
   inc nX
   savef nX
opend 2

;INY
opbegin _200
   inc nY
   savef nY
opend 2


;DEX
opbegin _202
   dec nX
   savef nX
opend 2

;DEY
opbegin _136
   dec nY
   savef nY
opend 2


.MACRO INC_op amode,op,cy
  opbegin &op
	R&amode
	inc	 al
	savef al
	W&amode
  opend &cy
.END

.MACRO DEC_op amode,op,cy
  opbegin &op
	R&amode
	dec	 al
	savef al
	W&amode
  opend &cy
.END

INC_op	zero ,_230,5
INC_op	zeroX,_246,6
INC_op	 abs ,_238,6
INC_op	 absX,_254,7

DEC_op	zero ,_198,5
DEC_op	zeroX,_214,6
DEC_op	 abs ,_206,6
DEC_op	 absX,_222,7


;----------------
; shift


;ASLA
opbegin _10
  shl nA,1
  savefc nA
opend 2

;LSRA
opbegin _74
  shr nA,1
  savefc nA
opend 2

;ROLA
opbegin _42
  getc
  rcl nA,1
  savefc nA
opend 2

;RORA
opbegin _106
  getc
  rcr nA,1
  savefc nA
opend 2


.MACRO ASL_op amode,op,cy
  opbegin &op
	R&amode
	shl	   al,1
	savefc al
	W&amode
  opend &cy
.END

.MACRO LSR_op amode,op,cy
  opbegin &op
	R&amode
	shr	   al,1
	savefc al
	W&amode
  opend &cy
.END


.MACRO ROL_op amode,op,cy
  opbegin &op
	R&amode
	getc
	rcl	   al,1
	savefc al
	W&amode
  opend &cy
.END

.MACRO ROR_op amode,op,cy
  opbegin &op
	R&amode
	getc
	rcr	   al,1
	savefc al
	W&amode
  opend &cy
.END

ASL_op	 zero,_6,5
ASL_op	zeroX,_22,6
ASL_op	  abs,_14,6
ASL_op	 absX,_30,7

LSR_op	 zero,_70,5
LSR_op	zeroX,_86,6
LSR_op	  abs,_78,6
LSR_op	 absX,_94,7

ROL_op	 zero,_38,5
ROL_op	zeroX,_54,6
ROL_op	  abs,_46,6
ROL_op	 absX,_62,7

ROR_op	 zero,_102,5
ROR_op	zeroX,_118,6
ROR_op	  abs,_110,6
ROR_op	 absX,_126,7



;----------
;flag

;CLC
opbegin _24
  mov BYTE [n6502cpu_carry], dl ; dl=0
opend 2

;CLD
opbegin _216
  and BYTE [n6502cpu_rF], NOT F_DEC
opend 2

;CLI
opbegin _88
  and BYTE [n6502cpu_rF], NOT F_INT
opend 2

;CLV
opbegin _184
  mov BYTE [n6502cpu_over], dl
opend 2

;SEC
opbegin _56
  mov BYTE [n6502cpu_carry], 01h
opend 2

;SED
opbegin _248
  or BYTE [n6502cpu_rF], F_DEC
opend 2

;SEI
opbegin _120
  or BYTE [n6502cpu_rF], F_INT
opend 2


;---------------
; Txx

;TAX
opbegin _170
   mov nX,nA
   savef nA
opend 2

;TXA
opbegin _138
   mov nA,nX
   savef nX
opend 2

;TAY
opbegin _168
   mov nY,nA
   savef nA
opend 2

;TYA
opbegin _152
   mov nA,nY
   savef nY
opend 2

;TSX
opbegin _186
   mov nX,nS
   savef nS
opend 2

;TXS
opbegin _154
   mov nS,nX
opend 2



;-----------------
; branch

;execute branch
quickbranch:
  fetchbytesx edx
  mov al,4		  ;4 cycles for branch taken  (should check page)
  add esi,edx
  jmp executeloop

dobranch:
  mov al,[esi]
  inc esi

  sub esi,[n6502cpu_PCBASE]
  test al,al

  mov edx,esi
  jl  brbackward

  ;forward branch
brforward:
  add dl,al ;add branch
  jnc short donebranch
  inc dh
  mov esi,edx
  mov al,4	 ;4 cycles
;  call naturalizeESI
  add esi,[n6502cpu_PCBASE]
  jmp executeloop

  ;backward branch
brbackward:
  add dl,al ;add branch
  jc  short donebranch
  dec dh
  mov esi,edx
  mov al,4	 ;4 cycles
;  call naturalizeESI
  add esi,[n6502cpu_PCBASE]
  jmp executeloop

  ;no page crossing
donebranch:
  mov esi,edx
  mov al,3	  ;3 cycles
;  call naturalizeESI
  add esi,[n6502cpu_PCBASE]
  jmp executeloop




;BPL rel
opbegin _16
  mov  dl,[n6502cpu_sign]

  test dl,dl
  jns dobranch
  inc esi
opend 2

;BMI rel
opbegin _48
  mov  dl,[n6502cpu_sign]

  test dl,dl
  js  dobranch
  inc esi
opend 2

;BVC rel
opbegin _80
  cmp dl,[n6502cpu_over]
  je  dobranch
  inc esi
opend 2

;BVS rel
opbegin _112
  cmp dl,[n6502cpu_over]
  jne  dobranch
  inc esi
opend 2


;BCC rel
opbegin _144
  cmp dl,[n6502cpu_carry]
  je  dobranch
  inc esi
opend 2

;BCS rel
opbegin _176
  cmp dl, [n6502cpu_carry]
  jne  dobranch
  inc esi
opend 2


;BNE rel
opbegin _208
  cmp dl, [n6502cpu_zero]
  jne dobranch
  inc esi
opend 2


;BEQ rel
opbegin _240
  cmp dl, [n6502cpu_zero]
  je  dobranch
  inc esi
opend 2


;---------------
;jump


;JMP $XXXX
opbegin _76
   movzx esi, word [esi]
   call naturalizeESI
opend 3


;JMP ($XXXX)
opbegin _108
   push ebx
   mov	dx, word [esi] ;get addr of address
   xor	ebx, ebx

   call readbyte
   mov	bl,al

   inc	dl

   call readbyte
   mov	bh,al

   mov	esi,ebx
   pop	ebx
   call naturalizeESI
opend	5

;JSR $XXXX
opbegin _32
   mov	 eax,esi  ;get 32-bit PC
   xor	 esi,esi

   mov	 dl,nS	  ;get SP
   mov	 si,[eax] ;get new PC

   sub	 eax,[n6502cpu_PCBASE] ;get 16-bit PC
   call	 naturalizeESI ;set new 32-bit PC

   inc ax			 ;increase old PC
   npushb ah
   npushb al

   mov nS,dl		 ;set SP
opendzx 6


;RTS
opbegin _96
  mov dl,nS ;get SP

  ;pop PC
  npopb al
  npopb ah

  ;store SP inc PC
  mov nS,dl
  inc ax

  ;set new PC
  mov  esi,eax
  call naturalizeESI
opendzx 6


;RTI
opbegin _64
  mov dl,nS ;get SP

  ;pop flags
  npopb al
  or  al,F_RSRVD
  degenerateflags

  ;pop PC
  npopb al
  npopb ah

  ;store SP
  mov nS,dl

  ;set new PC
  mov  esi,eax
  call naturalizeESI
opendzx 6



;-------------------
; stack


;PHA
opbegin _72
  mov dl,nS
  dec nS
  mov [ebp+edx+100h],nA
opend 3

;PHP
opbegin _8
  mov dl,nS
  dec nS
  generateflags
  mov [ebp+edx+100h],al
opend 3

;PLA
opbegin _104
  inc nS
  mov dl,nS
  mov nA,[ebp+edx+100h]
  savef nA
opend 4

;PLP
opbegin _40
  inc nS
  mov dl, nS
  mov al, BYTE [ebp+edx+100h]
  or  al, F_RSRVD
  degenerateflags
opend 4


;---------------------------
; BRK


opbegin _0
   denaturalizeESI
   mov eax, esi ;get pc

   mov dl, nS	;get stack
   inc ax

   ;push pc
   npushb ah
   npushb al

   ;push flags
   mov al, BYTE [n6502cpu_rF]
   or  al, F_BREAK ;set break flag ?
   npushb al

   ;store S
   mov nS, dl

   ;adjust flags
   or  al, F_INT ;turn on interrupt
   xor esi, esi
   mov BYTE [n6502cpu_rF], al

   ;PC=BRK vector
   mov si, word [edi+0FFFEh]
   call naturalizeESI
opendzx 7

;--------------------------------------

	.SECTION DATA

; ------- jump table ------------------
.MACRO makeopcode X
.if defined(op&X)
  dd op&X
.else
  dd badopcode
.end
.END

n6502jmptable:
.REPEAT 256
  makeopcode &count
.END

; --------- End Of File ----------------------------------------------

